/* MoarNES source - render.c

   This open-source software is (c)2010-2011 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to render NES frames to the screen, as well
   as managing the SDL screen context.
*/

#include <stdio.h>
#include "common.h"
#ifdef _WIN32
	#include "SDL/include/SDL_syswm.h"
#endif
#include "stdint.h"
#include "render.h"
#include "font.h"

SDL_Surface *screen;
uint16_t screenmode = 2, lastscreenmode = 2;

extern uint8_t outputNES[240][256];

#ifdef _WIN32
HWND myWindow;
HINSTANCE myInstance;
HMENU myMenu;
WNDPROC oldProc;
HWND GetHwnd();
HICON myIcon;
void SetWndProc();
#endif

bool initSDL(uint8_t *windowcaption) {
	if (SDL_Init(SDL_INIT_EVERYTHING) < 0) return(false);

	screen = SDL_SetVideoMode(512, 480, 32, SDL_HWSURFACE);
	if (screen == NULL) {
		SDL_Quit();
		return(false);
	}

	SDL_WM_SetCaption(windowcaption, NULL);
	#ifdef _WIN32
		myWindow = GetHwnd();
		myInstance = GetModuleHandle(NULL);
		myMenu = LoadMenu(NULL, MAKEINTRESOURCE(IDR_MENU1));
		SetMenu(myWindow, myMenu);
		SetWndProc();
		screen = SDL_SetVideoMode(512, 480, 32, SDL_HWSURFACE);
		if (screen == NULL) {
			SDL_Quit();
			return(false);
		}
		myIcon = LoadIcon(myInstance, MAKEINTRESOURCE(IDI_ICON1));
		SetClassLong(myWindow, GCLP_HICON, (LONG)(uint64_t)myIcon);
		SetClassLong(myWindow, GCLP_HICONSM, (LONG)(uint64_t)myIcon);
	#endif

	return(true);
}

int8_t *readSDLerr() {
	return(SDL_GetError());
}

#ifdef _WIN32
HWND GetHwnd() {
	SDL_SysWMinfo wmi;
	SDL_VERSION(&wmi.version);

	if (!SDL_GetWMInfo(&wmi)) return(NULL);
	return(wmi.window);
}

extern void MenuItemClick(WPARAM wParam);

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch (msg) {
		case WM_COMMAND:
			MenuItemClick(wParam);
			return(TRUE);
	}

	return(CallWindowProc(oldProc, hwnd, msg, wParam, lParam));
}

void SetWndProc() {
	oldProc = SetWindowLong(myWindow, GWL_WNDPROC, (LONG)WndProc);
}
#endif

void blitframe(uint8_t *NESbuffer) {
	uint32_t bufptr, SDLptr = 0;
	int32_t x, y, calcx, calcy;
	//float xscale, yscale;

	if (SDL_MUSTLOCK(screen)) 
		if (SDL_LockSurface(screen) < 0) 
			return;

	//xscale = (float)256.0 / (float)DISP_WIDTH;
	//yscale = (float)224.0 / (float)DISP_HEIGHT;
			//calcx = (uint32_t)(xscale * (float)x);
			//calcy = (uint32_t)(yscale * (float)y) + 8;

	switch (screenmode) {
		case 1:
			bufptr = 0;
			SDLptr = 0;
			for (y=0; y<screen->h; y++) {
				for (x=0; x<screen->w; x++) {
					((uint32_t *)screen->pixels)[SDLptr++] = paletteNES[NESbuffer[bufptr++]];
				}
			}
			break;

		case 2:
			bufptr = 0;
			SDLptr = 0;
			for (y=0; y<480; y+=2) {
				for (x=0; x<512; x++) {
					((uint32_t *)screen->pixels)[SDLptr] = paletteNES[NESbuffer[bufptr]];
					((uint32_t *)screen->pixels)[SDLptr++ + screen->w] = paletteNES[NESbuffer[bufptr]];
					if (x & 1) bufptr++;
				}
				SDLptr += screen->w;
			}
			break;

		case 3:
			for (y=0; y<screen->h; y++) {
				for (x=0; x<screen->w; x++) {
					calcx = x / 3;
					calcy = y / 3;
					bufptr = (calcy << 8) + calcx;
					SDLptr = (y * screen->w) + x;
					((uint32_t *)screen->pixels)[SDLptr] = paletteNES[NESbuffer[bufptr]];
				}
			}
			break;

		case 4:
			for (y=0; y<screen->h; y++) {
				for (x=0; x<screen->w; x++) {
					calcx = x >> 2;
					calcy = y >> 2;
					bufptr = (calcy << 8) + calcx;
					SDLptr = (y * screen->w) + x;
					((uint32_t *)screen->pixels)[SDLptr] = paletteNES[NESbuffer[bufptr]];
				}
			}
			break;

		case 5:
			bufptr = 0;
			for (y=0; y<480; y+=2) {
				SDLptr = (y * screen->w) + 64;
				for (x=0; x<512; x++) {
					((uint32_t *)screen->pixels)[SDLptr] = paletteNES[NESbuffer[bufptr]];
					((uint32_t *)screen->pixels)[SDLptr++ + screen->w] = paletteNES[NESbuffer[bufptr]];
					if (x & 1) bufptr++;
				}
			}
			break;
	}

  if (SDL_MUSTLOCK(screen)) 
    SDL_UnlockSurface(screen);

  SDL_UpdateRect(screen, 0, 0, screen->w, screen->h);
}

void setscreenmode(uint16_t modeval) {
	switch (modeval) {
		case 1:
			if (screenmode == 5) SetMenu(myWindow, myMenu);
			screen = SDL_SetVideoMode(256, 240, 32, SDL_HWSURFACE);
			lastscreenmode = screenmode;
			screenmode = 1;
			break;

		case 2:
			if (screenmode == 5) SetMenu(myWindow, myMenu);
			screen = SDL_SetVideoMode(512, 480, 32, SDL_HWSURFACE);
			lastscreenmode = screenmode;
			screenmode = 2;
			break;

		case 3:
			if (screenmode == 5) SetMenu(myWindow, myMenu);
			screen = SDL_SetVideoMode(768, 720, 32, SDL_HWSURFACE);
			lastscreenmode = screenmode;
			screenmode = 3;
			break;

		case 4:
			if (screenmode == 5) SetMenu(myWindow, myMenu);
			screen = SDL_SetVideoMode(1024, 960, 32, SDL_HWSURFACE);
			lastscreenmode = screenmode;
			screenmode = 4;
			break;

		case 5:
			SetMenu(myWindow, NULL);
			screen = SDL_SetVideoMode(640, 480, 32, SDL_HWSURFACE | SDL_FULLSCREEN);
			lastscreenmode = screenmode;
			screenmode = 5;
			break;
	}
	if (modeval == 5) SDL_ShowCursor(SDL_DISABLE);
		else SDL_ShowCursor(SDL_ENABLE);
	blitframe(&outputNES[0][0]);
}

void drawtext(uint8_t *text, uint16_t x, uint16_t y) {
	uint8_t curchar = 0;
	uint16_t textpos = 0, calcx, calcy;

	for (;;) {
		curchar = text[textpos++];
		if (curchar == 0) break;
		for (calcy=y; calcy<(y+8); calcy++) {
			for (calcx=x; calcx<(x+8); calcx++) {
				outputNES[calcy][calcx] = 65 - font[curchar][(calcy-y)*8+(calcx-x)];
			}
		}
		x += 8;
	}
}
